//
// This simple program approximates first- and second-order spatial derivatives using
// finite difference schemes.
// The function considered is exp(x) and we compute derivatives at x=1.0
// Order of convergence of the following schemes are studied:
// 1. Forward differece for first derivative
// 2. Backward differece for first derivative
// 3. 2nd order central differece for first derivative
// 4. 2nd order central differece for second derivative
//
// This is a part of teaching materials provided for the course
// "Computational Heat & Fluid Flow (ME 605)" taught at IIT Goa
// during the spring semester of 2021
//
// Written by: Dheeraj R, IIT Goa
// email: dheeraj@iitgoa.ac.in
//
// Tested with g++ 9.3.0 compiler on 09 February 2021
//
// Assignment for the students
// 1. second order one-sided difference for first derivative
// 2. Fourth order central difference approximation for first derivative
// 3. Fourth order central difference approximation for second derivative
//
// Instructions to run the code
// 1. compile this function and run -- this will create output file output.txt
// 2. execute the associated python code plot.py, which will read output.txt and produce
// the required plot in PDF format
//
//
//

#include <vector>
#include <math.h>
#include <fstream>			// writing files
#include <iomanip> 			// setprecision()

using namespace std;

int main() {
    int num_refine = 10;
    
    vector<double> h(num_refine);		// mesh spacing for each refinement
    vector<double> bd(num_refine);		// approximation by backward differencing
    vector<double> bderr(num_refine);		// error introduced in backward differencing
    vector<double> fd(num_refine);		// approximation by forward differencing
    vector<double> fderr(num_refine);		// error introduced in forward differencing
    vector<double> cd(num_refine);		// approximation by central differencing
    vector<double> cderr(num_refine);		// error introduced in central differencing
    vector<double> cd2(num_refine);		// approximation of second derivative by central differencing
    vector<double> cd2err(num_refine);	// error introduced in central differencing for second derivative
    
    
    // exact value at x=1.0
    double exac = exp(1.0);
    
    // compute error introduced in the approximation by all the schemes
    for (int i=0; i<num_refine; i++) {
        h[i]=0.1/pow(2,i);
        double xp=1.0;            		// location and solution value at the considered node
        double fp=exp(xp);
        double xw=1.0-h[i];      		// left side node (denoted with 'w' meaning 'west' as in finite volume context)
        double fw=exp(xw);
        double xe=1.0+h[i];       		// right side node (denoted with 'e' meaning 'east' as in finite volume context)
        double fe=exp(xe);
        bd[i]=(fp-fw)/h[i];         		// backward difference approximation and error
        bderr[i]=abs(exac-bd[i]);
        fd[i]=(fe-fp)/h[i];        		// forward difference approximation and error
        fderr[i]=abs(exac-fd[i]);
        cd[i]=0.5*(fe-fw)/h[i];     		// central difference approximation and error
        cderr[i]=abs(exac-cd[i]);
        cd2[i]=(fw+fe-2.0*fp)/h[i]/h[i];  	// second order derivatives using central differencing
        cd2err[i]=abs(exac-cd2[i]);
    }
    
    // Lines showing first and second order convergence
    vector<double> fir(num_refine);
    vector<double> sec(num_refine);
    vector<double> fou(num_refine);
    fir[0] = 0.05;
    sec[0] = 0.05;	// 0.01		
    fou[0] = 0.05;
    int i = 1;
    
    while (i < num_refine) {
        fir[i] = fir[i-1]*(h[i]/h[i-1]);
        sec[i] = sec[i-1]*pow((h[i]/h[i-1]),2.0);
        fou[i] = fou[i-1]*pow((h[i]/h[i-1]),4.0);
        i = i+1;
    }
    
    // Writing results to a file
    ofstream myfile;
    myfile.open("output.txt");
    for (int i=0; i<num_refine; i++) {
        myfile << fixed << setprecision(15) << h[i] << " " << bderr[i] << " " << fderr[i] << " " 
        << cderr[i] << " " << cd2err[i] << " " << fir[i] << " " 
        << sec[i] << " " << fou[i] << "\n";
    }
    myfile.close();   
	    
    return 0;
}

